/*
           *** cmdftp, command line ftp client ***

    Copyright (C) 2003-2004 Claudio Fontana

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (look for the file called COPYING);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    You can contact the author (Claudio Fontana) by sending a mail
    to sick_soul@users.sourceforge.net
	                                                                 */
#include "cmdftp.h"

/*******************************/
/* ERROR, WARNING, ... STRINGS */
/*******************************/

char* msg_err[] = {
  0, "bad server response", "unrecognized option:", "wrong option value:", "option without value:", "connection failed", "giving up on login", "unexpected error", "heap allocation failed"
};

char* msg_war[] = {
  "bye", "cannot open local file:", "error writing to local file:", "error reading from local file:", "transfer not confirmed", "file not found", "cannot open temp file:", "unknown command:", "missing argument", "ignoring argument", "login failed", "operation failed or incomplete:", "missing environment variable:", "Resuming connection lost during", "interrupted during"
};

#if HAVE_GETOPT_LONG

char* msg_usage = "cmdftp [option [option_value]]* hostname\n\n"
"options: \n"
" -h, --help       show this help and quit\n"
" -v, --version    show version and quit\n"
" -p, --port N     specify the remote TCP port of the server (def:21)\n"
" -b, --buffer N   specify the size in bytes of the transfer buffer (def:8192)\n"
" -a, --attempts N number of manual login attempts\n"
" -q, --quiet      be (very) quiet\n"
" -n, --no-auto    disable autologin (~/.netrc)\n"
" -m, --no-manual  disable manual login\n"
" -g, --no-paging  disable output paging\n"
" -d, --dotfiles   do not ignore dot-files in transfers\n";
#else
char* msg_usage = "cmdftp [option [option_value]]* hostname\n\n"
"options: \n"
" -h             show this help and quit\n"
" -v             show version and quit\n"
" -p N           specify the remote TCP port of the server (def:21)\n"
" -b N           specify the size in bytes of the transfer buffer (def:8192)\n"
" -a N           number of manual login attempts\n"
" -q             be (very) quiet\n"
" -n             disable autologin (~/.netrc)\n"
" -m             disable manual login\n"
" -g             disable output paging\n"
" -d             do not ignore dot-files in transfers\n";
#endif

char* msg_version = PACKAGE_STRING " Copyright (c) 2003-2004 Claudio Fontana";

char* msg_intro = 
"cmdftp comes with ABSOLUTELY NO WARRANTY; this is free software, and you\n"
"are welcome to redistribute it under certain conditions specified by the\n"
"GPL (GNU General Public License). Look at the COPYING file for the details.\n"
"\nType h after the login to get help about internal commands.\n";

char* msg_help[] = {
  "h", "display this help",
  "l", "switch to local mode, following commands refer to local",
  "r", "switch to remote mode, following commands refer to remote",
  "pwd", "prompt working directory",
  "cd PATH", "change working directory",
  "md PATH", "make directory",
  "rd PATH", "remove empty directory",
  "rm MASK", "delete regular files matching MASK, skip directories",
  "ls", "list current directory contents",
  "ls MASK", "list files/dirs matching MASK",
  "dir", "pretty list of current directory",
  "cp MASK DIR", "copy all files/dirs matching MASK to DIR (recurs)",
  "mv MASK DIR", "move all files/dirs matching MASK to DIR (recurs)",
  "u  MASK DIR", "upload files/dirs matching MASK into remote DIR (recurs)",
  "d  MASK DIR", "download files/dirs matching MASK into local DIR (recurs)",
  "dr MASK DIR", "same as above. If local file already exists, resume",
  "p  FILE", "print contents of the FILE on the terminal",
  "e  FILE", "edit FILE",
  "q", "quit client",
  "quit|exit|bye", "aliases for q command",
  "<TAB>", "tab-completion"
};


/***********/
/* GLOBALS */
/***********/

struct cmdftp_options o = { 0, 21, 8192, 1, 0, 0, 0, 0, 0 };
                                                      /* program options */

int cmdftp_control = 0;                               /* control connection */
int cmdftp_data = 0;                                  /* data connection */

char* user = 0; char* pass = 0;
char logging_in = 0;                                  /* user_logging_in? */
volatile int transfer_interrupted = TRAN_INTR_NO;     /* Why? */

struct hostent* server = 0;                           /* the remote host */
struct termios cmdftp_t;                              /* terminal data */

char* buffer;                                         /* in/out file transf */

char cmd_buffer[CMD_BUF_SIZE + 1];                    /* in/out commands */
char cmd_line[CMD_BUF_SIZE + 1];                      /* in     line sep */
char cmd_userinput[CMD_BUF_SIZE + 1];                 /* buf for term in */
char* cmd_ptr;                                        /* ->cmd_buffer */

char* env[N_ENV];                                     /* used env vars */
char* cwd[2] = { 0 };                                 /* working dir {l, r} */

int mode = CMDFTP_REMOTE;                             /* current mode {l, r} */

struct sigaction sa;

/*********************************************/
/* USER COMMANDS and corresponding functions */
/*********************************************/

char* commands[] = { "bye", "cd", "cp", "d", "dir", "dr", "e", "exit", "h", "l", "ls", "md", "mv", "p", "pwd", "q", "quit", "r", "rd", "rm", "u" };

#define N_COMMANDS (sizeof(commands) / sizeof(char*))

fun_ptr functions[2][N_COMMANDS] = {
  { &quit, &local_cd, &local_cp, &d, &local_dir, &dr, &local_e, &quit, &h, &local_mode, &local_ls, &local_md, &local_mv, &local_p, &local_pwd, &quit, &quit, &remote_mode, &local_rd, &local_rm, &u },
  { &quit, &remote_cd, &remote_cp, &d, &remote_dir, &dr, &remote_e, &quit, &h, &local_mode, &remote_ls, &remote_md, &remote_mv, &remote_p, &remote_pwd, &quit, &quit, &remote_mode, &remote_rd, &remote_rm, &u }
};


/************************************/
/* ERROR, WARNING, OUTPUT FUNCTIONS */
/************************************/


void cmdftp_err(int code, char* msg)
  { fprintf(stderr, "cmdftp: %s %s\n", msg_err[code], msg); exit(code); }

void cmdftp_war(int code, char* msg)
  { if (!o.q) fprintf(stderr, "cmdftp: %s %s\n", msg_war[code], msg); }

void usage(void) { printf("\n%s\n", msg_usage); }
void version(void) { printf("%s\n", msg_version); }

void intro(void) {
  int i; if (o.q) return;
  for (i = 0; i < 79; i++) putchar('*');
  putchar('\n');
  version();
  printf("\n%s\n", msg_intro);
  for (i = 0; i < 79; i++) putchar('*');
  putchar('\n');
}

void print_progress(off_t cur_pos, off_t total_size) {
  double progress = cur_pos * 100.0 / total_size;
  printf(progress < 1.0 ? "%4.2f%%\r" : "%4.1f%%\r", progress);
  fflush(stdout);
}

/***************************************/
/* MAIN and other high level functions */
/***************************************/


int main(int argc, char* argv[]) {

  if (argc < 2) { usage(); exit(-1); }
  getoptions(argc, argv);

  intro();
  reset_cmd_buffer(); buffer = my_malloc(o.b);

  tcgetattr(0, &cmdftp_t); atexit(&cmdftp_canon_mode);
  init_signals();

  if (!(cmdftp_control = cmdftp_connect(o.p)))
    cmdftp_err(CMDFTP_ERR_CONN, "");

  greeting();
  login_procedure();

  if ((cwd[CMDFTP_REMOTE] = remote_pwd_aux()))
    cwd[CMDFTP_REMOTE] = my_strdup(cwd[CMDFTP_REMOTE]);
  else cmdftp_err(CMDFTP_ERR_SERV, "(PWD)");

  cwd[CMDFTP_LOCAL] = my_strdup(local_pwd_aux());

  while (1) {
    char* cmd;
    if (!o.q) { printf("cmdftp:%s>", cwd[mode]); fflush(stdout); }
    if (!(dispatch(cmd = readline(1, 1)))) return 0; 
    free(cmd);
  }
}

#if HAVE_GETOPT_LONG

struct option options[] = { 
  { "help", 0, 0, 'h' },
  { "version", 0, 0, 'v' },
  { "attempts", 1, 0, 'a' },
  { "port", 1, 0, 'p' },
  { "buffer", 1, 0, 'b' },
  { "quiet", 1, 0, 'q' },
  { "no-manual", 0, 0, 'm' },
  { "no-auto", 0, 0, 'n' },
  { "no-paging", 0, 0, 'g' },
  { "dotfiles", 0, 0, 'd' },
  { 0, 0, 0, 0 }
};

#endif

void getoptions(int argc, char* argv[]) {
  int i, rv;

  while ((i =

#if HAVE_GETOPT_LONG
	  getopt_long(argc, argv, "a:b:dghmnp:qv", options, 0)
#else
	  getopt(argc, argv, "a:b:dghmnp:qv")
#endif
	  
	  ) != -1) {
    rv = 1;
    
    switch (i) {
    case 'h' : usage(); exit(0);
    case '?' : case ':' : usage(); exit(-1);
    case 'v' : version(); exit(0);

    case 'a' : rv = sscanf(optarg, "%hu", &o.a); break;
    case 'p' : rv = sscanf(optarg, "%hu", &o.p); break;
    case 'b' : rv = sscanf(optarg, "%hu", &o.b); break;

    case 'q' : o.q = 1; break;
    case 'm' : o.m = 1; break;
    case 'n' : o.n = 1; break;
    case 'g' : o.g = 1; break;
    case 'd' : o.d = 1; break;
    }
    if (rv != 1) cmdftp_err(CMDFTP_ERR_WOPT, optarg);
  }

  if (optind >= argc) { usage(); exit(-1); }
  o.hostname = my_strdup(argv[optind]);
  
  env[CMDFTP_ENV_PAGER] = o.g ? 0 : getenv("PAGER");
  env[CMDFTP_ENV_EDITOR] = getenv("EDITOR");
  env[CMDFTP_ENV_HOME] = getenv("HOME");
}

void greeting(void) {
  if (recv_answer(0, 0) != 220) cmdftp_err(CMDFTP_ERR_SERV, "(greeting)");
}


/*********************************/
/* LOGIN AND AUTOLOGIN FUNCTIONS */
/*********************************/

void login_procedure(void) {
  int attempt;
  user = pass = 0;

  if (!o.n) {
    if (auto_login()) return;
  }

  if (!o.m) {
    for (attempt = 0; attempt < o.a; attempt++) {
      if (user) free(user); if (pass) free(pass);
      user = pass = 0;
      if (manual_login()) return;
      cmdftp_war(CMDFTP_WAR_LGIN, "");
    }
  }

  cmdftp_err(CMDFTP_ERR_LGIN, "");
}

int login(char* user, char* pass) {
  int rv = 0;
  
  logging_in = 1;
  snprintf(cmd_buffer, CMD_BUF_SIZE, "USER %s", user);
  if (send_command(cmd_buffer) != 331) return 0;

  snprintf(cmd_buffer, CMD_BUF_SIZE, "PASS %s", pass);
  if (send_command(cmd_buffer) == 230) rv = 1;
  logging_in = 0;
  return rv;
}

int manual_login(void) {
  printf("Username: ");
  if (!(user = readline(0, 1))) return 0;

  cmdftp_pwd_start(); printf("Password: ");
  pass = readline(0, 0);
  cmdftp_pwd_end();

  if (!pass) return 0;
  return login(user, pass);
}

int auto_login(void) {
#define AUTO_LINE_LEN 160
  FILE* f; char* fname; char line[AUTO_LINE_LEN]; char token[AUTO_LINE_LEN];
  int state, rv;
  char* rc = ".netrc"; state = rv = 0;

  if (!env[CMDFTP_ENV_HOME]) return 0;
  fname = my_malloc(strlen(env[CMDFTP_ENV_HOME]) + 1 + strlen(rc) + 1);
  sprintf(fname, "%s/%s", env[CMDFTP_ENV_HOME], rc);
  f = fopen(fname, "r"); free(fname); if (!f) return 0;

  while((fgets(line, AUTO_LINE_LEN, f))) {
    char* ptr, *ptr_new, *ptr_tmp; size_t l;
    l = strlen(line); if (line[l - 1] == '\n') line[l - 1] = 0;
    if ((ptr = strchr(line, '#')))
      if (ptr == line || ptr[-1] != '\\') *ptr = 0;
    ptr = ptr_new = ptr_tmp = line;

  tokenize:
    if ((l = auto_login_next_token(&ptr, token)) == -1) continue;
    if (l == 0) {
      if (!state)
	{ if (strcmp(o.hostname, token) == 0) state++; goto tokenize; }
      break;
    } else if (l == 1) {
      if (!state) { state++; goto tokenize; }
      break;
    } else if (l == 2) {
      if (state && !user) user = my_strdup(token); goto tokenize;
    } else if (l == 3) {
      if (state && !pass) pass = my_strdup(token); goto tokenize;
    }
  } 
  if (user && !pass) pass = my_strdup("");
  if (user) rv = login(user, pass);
  if (!rv) {
    if (user) { free(user); user = 0; }
    if (pass) { free(pass); pass = 0; }
  }
  return rv;
#undef AUTO_LINE_LEN
}

int auto_login_next_token(char** hay, char* store) {
#define AUTO_N_KEYS 4
  char* keys[] = { "machine", "default", "login", "password" };
  int keys_len[] = { 7, 7, 5, 8 };
  char* ptr, *minptr; int i, min;

  min = -1; minptr = 0;

  for (i = 0; i < AUTO_N_KEYS; i++)
    if ((ptr = strstr(*hay, keys[i])))
      if (!minptr || ptr < minptr) { minptr = ptr; min = i; }

  if (min != -1) {
    *hay = minptr + keys_len[min];
    if (min != 1) {
      auto_login_read_token(hay, store);
    }
  }
  return min;
#undef AUTO_N_KEYS
}

void auto_login_read_token(char** hay, char* store) {
  int i, c;

  if (!isspace(**hay)) { *store = 0; return; }
  while (isspace(**hay)) (*hay)++;

  for (i = 0; ; i++) {
    c = **hay;
    if (!c || isspace(c)) break;

    if (c == '\\') {
      int n = 0; (*hay)++;
      if (!(c = **hay)) break;

      if (isdigit(c)) {
	sscanf(*hay, "%3o%n", (unsigned int*)&c, &n); *hay += n - 1;

      } else if (c == 'x') {
	(*hay)++; sscanf(*hay, "%2x%n", (unsigned int*)&c, &n); *hay += n - 1;

      } else {
	if (c == 'a') c = '\a'; else if (c == 'b') c = '\b';
	else if (c == 'f') c = '\f'; else if (c == 'n') c = '\n';
	else if (c == 't') c = '\t'; else if (c == 'v') c = '\v';
	else {
	/* get character as if no backslash present, good to escape spaces */
	}
      }
    }
    store[i] = c;
    (*hay)++;
  }

  store[i] = 0;
}


/***********************/
/* FUNCTION DISPATCHER */
/***********************/


int dispatch(char* cmd) {
  int idx;
  char* argv[2];

  if (!cmd) return 0;
  split_cmd(cmd, argv);

  if (*argv[0] == 0) return 1;		/* the empty, do nothing command */

  if ((idx = str_binsearch(argv[0])) == -1)
    cmdftp_war(CMDFTP_WAR_BADC, cmd);
  else if (!functions[mode][idx](argv[1]))
    cmdftp_war(CMDFTP_WAR_FAIL, cmd);
  return 1;
}


/*********************************/
/* DISPATCHED FUNCTIONS: general */
/*********************************/


int quit(char* arg) { 
  cmdftp_war(CMDFTP_WAR_BYE, "");
  exit(0);
}

int h(char* arg) { 
  int i;

  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  for (i = 0; i < (sizeof(msg_help) / sizeof(char*)); i += 2)
    printf("%-14s %s\n", msg_help[i], msg_help[i + 1]);
  return 1;
}
    
int local_mode(char* arg) {
  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  mode = CMDFTP_LOCAL;
  if (!o.q) printf("Current mode: local file system\n"); return 1;
}

int remote_mode(char* arg) {
  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  mode = CMDFTP_REMOTE;
  if (!o.q) printf("Current mode: remote file system\n"); return 1;
}


/************************************/
/* DISPATCHED FUNCTIONS: local mode */
/************************************/


int local_pwd(char* arg) {
  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  printf("%s\n", cwd[CMDFTP_LOCAL]);
  return 1;
}

int local_cd(char* arg) {
  int rv = 0;
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (local_cd_aux(arg)) rv = 1;
  free(cwd[CMDFTP_LOCAL]); cwd[CMDFTP_LOCAL] = my_strdup(local_pwd_aux());
  return rv;
}

int local_cd_aux(char* arg) { return (chdir(arg) == 0); }

int local_cp(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if ((!local_cd_aux(paths[1])) && (local_cd_aux(paths[0]))) {
    source_mask = fullpath(paths[0], "*");
  } else source_mask = my_strdup(paths[0]);

  local_cd_aux(cwd[CMDFTP_LOCAL]);
  rv = local_cp_aux(paths[1], source_mask);
  free(source_mask);
  return rv;
}

int local_cp_aux(char* target_dir, char* source_mask) {
  struct list_data d;
  char* target[3];
  int i, rv = 1;

  if (!local_cd_aux(target_dir)) local_md(target_dir);
  else local_cd_aux(cwd[CMDFTP_LOCAL]);

  init_list(&d);

  if (!local_fetch_list(source_mask, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  canonized_fn(target, target_dir);

  for (i = 0; i < d.count; i++) {
    struct stat si;
    if (stat(d.data[i].fullname, &si) != -1) {

      if (S_ISREG(si.st_mode)) {
	if (!local_copy(target[2], d.data + i)) { rv = 0; break; }

      } else if (S_ISDIR(si.st_mode)) {
	char* tmp_src, *tmp_trg; int subrv = 0;

	tmp_trg = fullpath(target[2], d.data[i].basename);
	tmp_src = fullpath(d.data[i].fullname, "*");
	subrv = local_cp_aux(tmp_trg, tmp_src);
	free(tmp_trg); free(tmp_src);
	if (!subrv) { rv = 0; break; }
      }
    }
  }
  free_list(&d); free_fn(target); 
  return rv;
}

int local_mv(char* arg) {
  char* paths[2]; struct list_data d;
  char* target[3];
  int i, rv = 1;

  split_cmd(arg, paths);

  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }

  init_list(&d);
  if (!local_fetch_list(paths[0], &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, paths[0]); return 1; }

  canonized_fn(target, paths[1]);

  for (i = 0; i < d.count; i++) {
    struct stat si;

    if (stat(d.data[i].fullname, &si) != -1) {
      if (S_ISREG(si.st_mode) || S_ISDIR(si.st_mode)) {
	char* fn;
	if (!local_md(target[2])) fn = fullpath(target[2], d.data[i].basename);
	else fn = my_strdup(target[2]);
	if (rename(d.data[i].fullname, fn) == -1) { free(fn); rv = 0; break; }
	free(fn);
      }
    }
  }
  free_list(&d); free_fn(target); 
  return rv;
}

int local_copy(char* target_dir, struct list_entry* source) {
  FILE* t, *s;
  char* target;
  int c, rv = 0;
  
  if (!(s = fopen(source->fullname, "rb"))) return 0;
  target = fullpath(target_dir, source->basename);
  
  if ((t = fopen(target, "wb"))) {
    while ((c = getc(s)) != EOF) putc(c, t);
    rv = 1;
  }

  free(target); if (t) fclose(t); fclose(s); return rv;
}

int u(char* arg) {
  char* paths[2], *source_mask;
  int rv = 0;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!remote_cd_aux(paths[1])) && (local_cd_aux(paths[0]))) {
    local_cd_aux(cwd[CMDFTP_LOCAL]);
    source_mask = fullpath(paths[0], "*");
  } else source_mask = my_strdup(paths[0]);

  remote_cd_aux(cwd[CMDFTP_REMOTE]);
  rv = u_aux(paths[1], source_mask);
  free(source_mask);
  return rv;
}

int u_aux(char* target_dir, char* source_mask) {
  struct list_data ld;
  char* target[3];
  int i, rv = 1;

  if (!remote_cd_aux(target_dir)) remote_md(target_dir);
  else remote_cd_aux(cwd[CMDFTP_REMOTE]);
  
  init_list(&ld);

  if (!local_fetch_list(source_mask, &ld)) return 0;
  if (ld.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  canonized_fn(target, target_dir);
  
  for (i = 0; i < ld.count; i++) {
    struct stat si;
    if (stat(ld.data[i].fullname, &si) != -1) {
      if (S_ISREG(si.st_mode)) {
	FILE* f;

	if (!(f = fopen(ld.data[i].fullname, "rb")))
	  cmdftp_war(CMDFTP_WAR_OPEN, ld.data[i].fullname);
	else {
	  char* fn; int up;
	  fn = fullpath(target[2], ld.data[i].basename);
	  up = upload(fn, f);
	  fclose(f); free(fn);
	  if (!up) { rv = 0; break; }
	}
      } else if (S_ISDIR(si.st_mode)) {
	char* tmp_src, *tmp_trg; int subrv = 0;

	tmp_src = fullpath(ld.data[i].fullname, "*");
	tmp_trg = fullpath(target[2], ld.data[i].basename);

	subrv = u_aux(tmp_trg, tmp_src);
	free(tmp_trg); free(tmp_src);
	if (!subrv) { rv = 0; break; }
      }
    }
  }
  free_list(&ld); free_fn(target);
  return rv;
}

int upload(char* target, FILE* source) {
  off_t start_pos, total_size;
  int port, rv;

  rv = 0; 

  if (!send_command("TYPE I")) return 0;

 start_transfer:

  fseeko(source, 0, SEEK_END); total_size = ftello(source);
  fseeko(source, 0, SEEK_SET);

  start_pos = 0;

  if (!(port = getport())) return 0;
  if (!(cmdftp_data = cmdftp_connect(port))) return 0;

  snprintf(cmd_buffer, CMD_BUF_SIZE, "STOR %s", target);
  if (send_command(cmd_buffer) != 150) { close(cmdftp_data); return 0; }
  
  if (!o.q)
    printf("Uploading %s, " OFF_FMT " bytes.\n", clean_fn(target), total_size);

  transfer_interrupted = TRAN_INTR_NO;

  while (start_pos < total_size) {
    size_t toread, bytes; ssize_t written;
    if (!o.q) print_progress(start_pos, total_size);
    toread = ((total_size - start_pos) < o.b) ? total_size - start_pos : o.b;

    if (!(bytes = fread(buffer, 1, toread, source))) break;
    if ((written = my_raw_write(buffer, bytes, cmdftp_data)) != bytes) break;

    start_pos += bytes;
  }

  close(cmdftp_data); 

  if (start_pos < total_size) {
    if (ferror(source)) cmdftp_war(CMDFTP_WAR_READ, basename(target));
    else if TRANSFER_INTERRUPTED_CHECK("data transfer", 1);
    else { cmdftp_err(CMDFTP_ERR_UNER, strerror(errno)); }
  } else {
    if (!recv_confirm()) cmdftp_war(CMDFTP_WAR_NOCT, "");
    rv = 1;
  }

  return rv;
}

int local_ls(char* arg) {
  struct list_data d;
  int rv;

  init_list(&d);
  if (!local_fetch_list(arg, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  rv = ls(&d); free_list(&d); return rv;
}

int local_dir(char* arg) {
  struct list_data d;
  int rv;

  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  init_list(&d);
  if (!local_fetch_pretty_list(&d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, ""); return 1; }

  rv = ls(&d); free_list(&d); return rv;
}

int local_md(char* arg) {
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (mkdir(arg, 0777) < 0) return 0;
  return 1;
}

int local_rd(char* arg) { 
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (rmdir(arg) < 0) return 0;
  return 1;
}

int local_rm(char* arg) {
  struct list_data d;
  int i, rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }

  init_list(&d);
  if (!local_fetch_list(arg, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  for (i = 0; i < d.count; i++) {
    struct stat si;
    if ((stat(d.data[i].fullname, &si) != -1) && (S_ISREG(si.st_mode)))
      if (unlink(d.data[i].fullname) == -1) break;
    if (i == d.count - 1) rv = 1;
  }
  free_list(&d);
  return rv;
}

int local_p(char* arg) {

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }

  if (env[CMDFTP_ENV_PAGER])
    return cmdftp_execute(env[CMDFTP_ENV_PAGER], arg, -1, -1);
  else {
    FILE* f; int c;
    if (!(f = fopen(arg, "r"))) return 0;
    while ((c = getc(f)) != EOF) putchar(c); printf("\n");
    fclose(f);
  }
  return 1;
}

int local_e(char* arg) {
  
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (!env[CMDFTP_ENV_EDITOR])
    { cmdftp_war(CMDFTP_WAR_MENV, "EDITOR"); return 0; }

  return cmdftp_execute(env[CMDFTP_ENV_EDITOR], arg, -1, -1);
}


/*************************************/
/* DISPATCHED FUNCTIONS: remote mode */
/*************************************/

int remote_pwd(char* arg) {
  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  printf("%s\n", cwd[CMDFTP_REMOTE]);
  return 1;
}

int remote_cd(char* arg) {
  int rv = 0;
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (remote_cd_aux(arg)) rv = 1;

  free(cwd[CMDFTP_REMOTE]);
  
  if ((cwd[CMDFTP_REMOTE] = remote_pwd_aux()))
    cwd[CMDFTP_REMOTE] = my_strdup(cwd[CMDFTP_REMOTE]);
  else if ((cwd[CMDFTP_REMOTE] = remote_pwd_aux()))
    cwd[CMDFTP_REMOTE] = my_strdup(cwd[CMDFTP_REMOTE]);
  else cmdftp_err(CMDFTP_ERR_SERV, "(PWD)");
				
  return rv;
}

int remote_cd_aux(char* arg) {
  snprintf(cmd_buffer, CMD_BUF_SIZE, "CWD %s", arg);
  return (send_command(cmd_buffer) == 250);
}

int d(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!local_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    remote_cd_aux(cwd[CMDFTP_REMOTE]);
    source_mask = fullpath(paths[0], "*");
  } else source_mask = my_strdup(paths[0]);

  local_cd_aux(cwd[CMDFTP_LOCAL]);
  rv = d_aux(paths[1], source_mask, 0);
  free(source_mask);
  return rv;
}

int dr(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!local_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    remote_cd_aux(cwd[CMDFTP_REMOTE]);
    source_mask = fullpath(paths[0], "*");
  } else source_mask = my_strdup(paths[0]);

  local_cd_aux(cwd[CMDFTP_LOCAL]);
  rv = d_aux(paths[1], source_mask, 1);
  free(source_mask);
  return rv;
}

int d_aux(char* target_dir, char* source_mask, int resume) {
  struct list_data ld;
  char* target[3];
  int i, rv = 1;

  if (!local_cd_aux(target_dir)) local_md(target_dir);
  local_cd_aux(cwd[CMDFTP_LOCAL]);

  init_list(&ld);

  if (!remote_fetch_list(source_mask, &ld)) return 0;
  if (ld.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  canonized_fn(target, target_dir);
  
  for (i = 0; i < ld.count; i++) {
    if (!remote_cd_aux(ld.data[i].fullname)) {
      char* fn; FILE* f;
      
      fn = fullpath(target[2], ld.data[i].basename);
      if (!(f = fopen(fn, resume ? "ab" : "wb")))
	cmdftp_war(CMDFTP_WAR_OPEN, fn);
      else {
	int dl = download(f, ld.data[i].fullname);
	fclose(f);
	if (!dl) { rv = 0; free(fn); break; }
      }
      free(fn);
    } else {
      char* tmp_src, *tmp_trg; int subrv = 0;

      remote_cd_aux(cwd[CMDFTP_REMOTE]); 
      tmp_src = fullpath(ld.data[i].fullname, "*");
      tmp_trg = fullpath(target[2], ld.data[i].basename);
      subrv = d_aux(tmp_trg, tmp_src, resume);
      free(tmp_trg); free(tmp_src);
      if (!subrv) { rv = 0; break; }
    }
  }

  free_list(&ld); free_fn(target);
  return rv;
}

int remote_cp(char* arg) {
  char* paths[2], *source_mask;
  int rv;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!remote_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    source_mask = fullpath(paths[0], "*");
  } else source_mask = my_strdup(paths[0]);

  remote_cd_aux(cwd[CMDFTP_REMOTE]);
  rv = remote_cp_aux(paths[1], source_mask);
  free(source_mask);
  return rv;
}

int remote_cp_aux(char* target_dir, char* source_mask) {
  struct list_data d;
  char* target[3];
  int i, rv = 1;

  init_list(&d);

  if (!remote_fetch_list(source_mask, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  if (!remote_cd_aux(target_dir)) remote_md(target_dir);
  else remote_cd_aux(cwd[CMDFTP_REMOTE]);

  canonized_fn(target, target_dir);
  
  for (i = 0; i < d.count; i++) {
    if (!remote_cd_aux(d.data[i].fullname)) {
      if (!remote_copy(target[2], d.data + i)) { rv = 0; break; }
    } else {
      char* tmp_src, *tmp_trg; int subrv = 0;
      
      remote_cd_aux(cwd[CMDFTP_REMOTE]);
      tmp_src = fullpath(d.data[i].fullname, "*");
      tmp_trg = fullpath(target[2], d.data[i].basename);
      subrv = remote_cp_aux(tmp_trg, tmp_src);
      free(tmp_src); free(tmp_trg);
      if (!subrv) { rv = 0; break; }
    }
  }

  free_list(&d); free_fn(target);
  return rv;
}

int remote_mv(char* arg) {
  char* paths[2], *source_mask;
  int rv, rd = 0;

  split_cmd(arg, paths);
  if ((*paths[0] == 0) || (*paths[1] == 0))
    { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  
  if ((!remote_cd_aux(paths[1])) && (remote_cd_aux(paths[0]))) {
    rd = 1;
    source_mask = fullpath(paths[0], "*");
  } else source_mask = my_strdup(paths[0]);

  remote_cd_aux(cwd[CMDFTP_REMOTE]);
  rv = remote_mv_aux(paths[1], source_mask);
  if (rd) remote_rd(paths[0]);
  free(source_mask);
  return rv;
}

int remote_mv_aux(char* target_dir, char* source_mask) {
  struct list_data d;
  char* target[3];
  int i, rv = 1;

  init_list(&d);

  if (!remote_fetch_list(source_mask, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, source_mask); return 1; }

  if (!remote_cd_aux(target_dir)) remote_md(target_dir);
  else remote_cd_aux(cwd[CMDFTP_REMOTE]);

  canonized_fn(target, target_dir);
  
  for (i = 0; i < d.count; i++) {
    if (!remote_cd_aux(d.data[i].fullname)) {

      snprintf(cmd_buffer, CMD_BUF_SIZE, "RNFR %s", d.data[i].fullname);

      if (send_command(cmd_buffer) != 350)
	{ cmdftp_war(CMDFTP_WAR_MASK, d.data[i].fullname); goto end_proc; }
      
      snprintf(cmd_buffer, CMD_BUF_SIZE, "RNTO %s/%s",
	       target[2], d.data[i].basename);

      if (send_command(cmd_buffer) != 250) { rv = 0; goto end_proc; }

    } else {
      char* tmp_src, *tmp_trg; int subrv = 0;
      
      remote_cd_aux(cwd[CMDFTP_REMOTE]);
      tmp_src = fullpath(d.data[i].fullname, "*");
      tmp_trg = fullpath(target[2], d.data[i].basename);
      subrv = remote_mv_aux(tmp_trg, tmp_src);
      free(tmp_src); free(tmp_trg);
      if (subrv) remote_rd(d.data[i].fullname);
      else { rv = 0; break; }
    }
  }

 end_proc:

  free_list(&d); free_fn(target);
  return rv;
}

int remote_copy(char* target_dir, struct list_entry* source) {
  
  FILE* t;
  char* t_fn, *target = 0;
  size_t len, rv = 0;

  if (!(t = cmdftp_temp(&t_fn)))
    { cmdftp_war(CMDFTP_WAR_TEMP, t_fn); return 0; }

  if (!download(t, source->fullname)) goto end_proc;

  len = strlen(target_dir);
  if (target_dir[len - 1] == '/') target_dir[(--len)] = 0;

  target = fullpath(target_dir, source->basename);
  if (upload(target, t)) rv = 1;

 end_proc:
  fclose(t); unlink(t_fn); free(t_fn);
  if (target) free(target); return rv;
}

int download(FILE* target, char* source) {
  off_t start_pos, total_size;
  int answer, port, rv;
  
  rv = 0;
  
  if (!send_command("TYPE I")) return 0;
  if ((total_size = getsize(source)) < 0) return 0;

 start_transfer:

  start_pos = ftello(target);

  snprintf(cmd_buffer, CMD_BUF_SIZE, "REST " OFF_FMT, start_pos);
  if (!(answer = send_command(cmd_buffer))) return 0;

  if (start_pos > 0 && answer != 350) {
    fseeko(target, 0, SEEK_SET); start_pos = 0;
  }

  if (!(port = getport())) return 0;
  if (!(cmdftp_data = cmdftp_connect(port))) return 0;

  snprintf(cmd_buffer, CMD_BUF_SIZE, "RETR %s", source);
  if (send_command(cmd_buffer) != 150) { close(cmdftp_data); return 0; }

  if (!o.q) {
    printf("Downloading %s, " OFF_FMT " bytes", source, total_size);
    if (start_pos > 0) {
      printf(", continuing at position " OFF_FMT, start_pos);
    }
    printf(".\n");
  }

  transfer_interrupted = TRAN_INTR_NO;

  while (start_pos < total_size) {
    ssize_t bytes; size_t toread, written;
    if (!o.q) print_progress(start_pos, total_size);

    toread = ((total_size - start_pos) < o.b) ? total_size - start_pos : o.b;

    if ((bytes = my_raw_read(buffer, toread, cmdftp_data)) <= 0) break;
    if ((written = fwrite(buffer, 1, bytes, target)) != bytes) break;

    start_pos += bytes;
  }

  close(cmdftp_data); 

  if (start_pos < total_size) {
    if (ferror(target))
      cmdftp_war(CMDFTP_WAR_WRIT, basename(source));
    else if TRANSFER_INTERRUPTED_CHECK("data transfer", 1);
    else { cmdftp_err(CMDFTP_ERR_UNER, strerror(errno)); }
  } else {
    if (!recv_confirm()) cmdftp_war(CMDFTP_WAR_NOCT, "");
    rv = 1;
  }

  return rv;
}

int ls(struct list_data* d) {
  FILE* target; int pipe_fd[2];
  int i, rv = 0;

  if (env[CMDFTP_ENV_PAGER]) {
    if (pipe(pipe_fd) < 0) return 0;
    if (!(target = fdopen(pipe_fd[1], "w")))
      { close(pipe_fd[0]); close(pipe_fd[1]); return 0; }
  } else target = stdout;
  
  for (i = 0; i < d->count; i++)
    fprintf(target, "%s\n", d->data[i].fullname);

  if (!env[CMDFTP_ENV_PAGER]) return 1;

  fclose(target);
  rv = cmdftp_execute(env[CMDFTP_ENV_PAGER], 0, pipe_fd[0], -1);

  close(pipe_fd[0]);
  return rv;
}

int remote_ls(char* arg) {
  struct list_data d;
  int rv;

  init_list(&d);
  if (!remote_fetch_list(arg, &d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  rv = ls(&d); free_list(&d); return rv;
}

int remote_dir(char* arg) {
  struct list_data d;
  int rv;

  if (*arg != 0) cmdftp_war(CMDFTP_WAR_IARG, arg);

  init_list(&d);
  if (!remote_fetch_pretty_list(&d)) return 0;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  rv = ls(&d); free_list(&d); return rv;
}

int remote_md(char* arg) {
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; } 
  snprintf(cmd_buffer, CMD_BUF_SIZE, "MKD %s", arg);
  if (send_command(cmd_buffer) != 257) return 0;
  return 1;
}

int remote_rd(char* arg) {
  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; } 
  snprintf(cmd_buffer, CMD_BUF_SIZE, "RMD %s", arg);
  if (send_command(cmd_buffer) != 250) return 0;
  return 1;
}

int remote_rm(char* arg) {
  struct list_data d;
  int i, rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  init_list(&d);
  if (!remote_fetch_list(arg, &d)) goto end_proc;
  if (d.count == 0) { cmdftp_war(CMDFTP_WAR_MASK, arg); return 1; }

  for (i = 0; i < d.count; i++) {
    snprintf(cmd_buffer, CMD_BUF_SIZE, "DELE %s", d.data[i].fullname);
    if (!send_command(cmd_buffer)) goto end_proc;
  }
  rv = 1;
  
 end_proc:
  free_list(&d);
  return rv;
}

int remote_p(char* arg) {
  int c; char* fn; FILE* f; int rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (!(f = cmdftp_temp(&fn))) { cmdftp_war(CMDFTP_WAR_TEMP, fn); return 0; }

  if (download(f, arg)) {
    if (env[CMDFTP_ENV_PAGER]) {
      fclose(f); f = 0;
      rv = cmdftp_execute(env[CMDFTP_ENV_PAGER], fn, -1, -1);
    } else {
      fseeko(f, 0, SEEK_SET);
      while ((c = getc(f)) != EOF) putchar(c); printf("\n");
      rv = 1;
    }
  }

  if (f) fclose(f); unlink(fn); free(fn); return rv;
}

int remote_e(char* arg) {
  char* fn; FILE* f; int rv = 0;

  if (*arg == 0) { cmdftp_war(CMDFTP_WAR_MARG, ""); return 0; }
  if (!env[CMDFTP_ENV_EDITOR])
    { cmdftp_war(CMDFTP_WAR_MENV, "EDITOR"); return 0; }

  if (!(f = cmdftp_temp(&fn))) { cmdftp_war(CMDFTP_WAR_TEMP, fn); return 0; }

  if (download(f, arg)) {
    fclose(f); f = 0;
    if (cmdftp_execute(env[CMDFTP_ENV_EDITOR], fn, -1, -1)) {
      if ((f = fopen(fn, "r")) && upload(arg, f)) rv = 1;
    }
  }
  if (f) fclose(f); unlink(fn); free(fn); return rv;
}


/******************************/
/* OTHER AUX FUNCTIONS: local */
/******************************/


char* local_pwd_aux(void) { return getcwd(cmd_buffer, CMD_BUF_SIZE); }

int local_fetch_list(char* filemask, struct list_data* d) {
  return local_fetch_list_aux(filemask, d, 0);
}

int local_fetch_pretty_list(struct list_data* d) {
  return local_fetch_list_aux(0, d, 1);
}

int local_fetch_list_aux(char* filemask, struct list_data* d, char pretty) {
  char* fn_mask[3], *tmpmask;
  struct dirent* entry;
  DIR* dir;
  int rv = 0;
  
  if (filemask) {
    if (*filemask == 0) tmpmask = my_strdup("*");
    else tmpmask = my_strdup(filemask);

    canonized_fn(fn_mask, tmpmask);
  }

  while (pretty) {
    int pipe_fd[2]; FILE* target; char line[CMD_BUF_SIZE];
    if (pipe(pipe_fd) < 0) break;

    if (!cmdftp_execute("/bin/ls -l", 0, -1, pipe_fd[1]))
      { close(pipe_fd[0]); close(pipe_fd[1]); break; }
    close(pipe_fd[1]);
    if (!(target = fdopen(pipe_fd[0], "r"))) { close(pipe_fd[0]); break; }
    while (fgets(line, CMD_BUF_SIZE, target)) {
      size_t len = strlen(line);
      if (line[len - 1] == '\n') line[len - 1] = 0;
      store_pretty_list(line, d);
    }
    fclose(target); rv = 1;
    goto end_proc;
  }

  if (!(dir = opendir(filemask ? fn_mask[0] : "."))) goto end_proc;

  while ((entry = readdir(dir))) {
    if (!filemask || fnmatch(fn_mask[1], entry->d_name, 0) == 0) {
      char* full_name = fullpath(filemask ? fn_mask[0] : ".", entry->d_name);
      store_list(clean_fn(full_name), d);
      free(full_name);
    }
  }

  closedir(dir); rv = 1;

 end_proc:
  if (filemask) { free(tmpmask); free_fn(fn_mask); }
  return rv;
}


/*******************************/
/* OTHER AUX FUNCTIONS: remote */
/*******************************/


char* remote_pwd_aux(void) {
  char* start, *end;
  if (!send_command("PWD")) return 0;
  if ( ((end = strrchr(cmd_buffer, '"')) == 0) ||
       ((start = strchr(cmd_buffer, '"')) == 0) ) return 0;
  *end = 0; start++;
  return start;
}

int remote_fetch_list(char* filemask, struct list_data* d) {
  return remote_fetch_list_aux(filemask, d, 0);
}

int remote_fetch_pretty_list(struct list_data* d) {
  return remote_fetch_list_aux(0, d, 1);
}

int remote_fetch_list_aux(char* filemask, struct list_data* d, char pretty) {
  char* tmpmask, *fn_mask[3], *line;
  int tmp, rv = 0;
  int port;

  if (!send_command("TYPE A")) return 0;

  if (filemask) {
    if (*filemask == 0) tmpmask = my_strdup("*");
    else tmpmask = my_strdup(filemask);

    canonized_fn(fn_mask, tmpmask);
  }
  
 start_transfer:
 
  if (!(port = getport())) goto end_proc;
  if (!(cmdftp_data = cmdftp_connect(port))) goto end_proc;

  if (filemask) {
    snprintf(cmd_buffer, CMD_BUF_SIZE, pretty ? "LIST %s" : "NLST %s",
	     fn_mask[0]);
  } else {
    sprintf(cmd_buffer, pretty ? "LIST" : "NLST");
  }

  if ((tmp = (send_command(cmd_buffer))) != 150)
    { close(cmdftp_data); if (tmp == 450) rv = 1; goto end_proc; }

  reset_cmd_buffer(); 

  transfer_interrupted = TRAN_INTR_NO;

  while ((line = recv_line(cmdftp_data))) {
    if (!pretty) {
      char* fn;
      if (!(fn = strrchr(line, '/'))) fn = line;
      else fn++;

      if (!filemask || fnmatch(fn_mask[1], fn, 0) == 0) {
	char* full_name = filemask ? fullpath(fn_mask[0], fn) : fn;
	store_list(clean_fn(full_name), d);
	if (filemask) free(full_name);
      }
    } else {
      store_pretty_list(line, d);
    }
  }

  close(cmdftp_data); 
  
  if TRANSFER_INTERRUPTED_CHECK("data transfer", 1);
  else {
    if (!recv_confirm()) cmdftp_war(CMDFTP_WAR_NOCT, "");
    rv = 1;
  }
  
 end_proc:
  if (filemask) { free(tmpmask); free_fn(fn_mask); }
  return rv;
}  


/*****************************/
/* NETWORK RELATED FUNCTIONS */
/*****************************/


int cmdftp_connect(int port) {
  struct sockaddr_in address; int s, opvalue; socklen_t slen;
  
  opvalue = 8; slen = sizeof(opvalue);
  memset(&address, 0, sizeof(address));

  if ( ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) ||
       ((setsockopt(s, IPPROTO_IP, IP_TOS, &opvalue, slen)) < 0) )
    return 0;

  address.sin_family = AF_INET;
  address.sin_port = htons((unsigned short)port);

  if (!server)
    if (!(server = gethostbyname(o.hostname))) return 0;

  memcpy(&address.sin_addr.s_addr, server->h_addr, server->h_length);
  
  if (connect(s, (struct sockaddr*) &address, sizeof(address)) == -1)
    return 0;

  return s;
}

void cmdftp_reconnect(void) {
  if (!(cmdftp_control = cmdftp_connect(o.p)))
    cmdftp_err(CMDFTP_ERR_CONN, "");
  greeting();
  if (!logging_in) {
    if (!login(user, pass)) cmdftp_err(CMDFTP_ERR_LGIN, "");
    if (cwd[CMDFTP_REMOTE])
      remote_cd(cwd[CMDFTP_REMOTE]);
  }
}

ssize_t my_raw_read(char* buf, size_t n, int sc) { 
  ssize_t rv;

  rv = recv(sc, buf, n, 0);

  if (rv < 0) {
    if (transfer_interrupted != TRAN_INTR_INT)
      transfer_interrupted = TRAN_INTR_PIPE;
  }
  return rv;
}

ssize_t my_raw_write(char* buf, size_t n, int sc) {
  ssize_t rv;

  rv = send(sc, buf, n, 0);

  if (rv < 0) {
    if (transfer_interrupted != TRAN_INTR_INT)
      transfer_interrupted = TRAN_INTR_PIPE;
  }
  return rv;
}

int send_command(char* cmd) { 
  char* s; size_t len, rv = 0;

  s = my_malloc((len = strlen(cmd) + 2) + 1);
  sprintf(s, "%s\r\n", cmd);

 start_transfer:
  transfer_interrupted = TRAN_INTR_NO;

  if (my_raw_write(s, len, cmdftp_control) < 0) {
    s[len - 2] = 0;
    if TRANSFER_INTERRUPTED_CHECK(s, 0);
  } else {
    s[len - 2] = 0;
    if (!(rv = recv_answer(0, 0))) {
      if TRANSFER_INTERRUPTED_CHECK(s, 1);
    }
  }
  free(s);
  return rv;
}

void reset_cmd_buffer(void) {
  cmd_ptr = cmd_buffer + CMD_BUF_SIZE;
}

char* recv_line(int sc) {
  char* new_ptr;
  int n; size_t len;

  if (*cmd_ptr == 0) {
    ssize_t bytes = my_raw_read(cmd_buffer, CMD_BUF_SIZE, sc);
    if (bytes <= 0) return 0;
    cmd_buffer[bytes] = 0;
    cmd_ptr = cmd_buffer;
    if (*cmd_ptr == '\n' && cmd_buffer[CMD_BUF_SIZE - 1] == '\r') cmd_ptr++;
  }

  new_ptr = strchr(cmd_ptr, '\r');
  if (!new_ptr) new_ptr = cmd_buffer + CMD_BUF_SIZE;

  n = new_ptr - cmd_ptr;
  strncpy(cmd_line, cmd_ptr, n); cmd_line[n] = 0;
  cmd_ptr = new_ptr;

  if (cmd_ptr == cmd_buffer + CMD_BUF_SIZE) {
    ssize_t bytes = my_raw_read(cmd_buffer, CMD_BUF_SIZE, sc);
    if (bytes <= 0) return 0;
    cmd_buffer[bytes] = 0;

    cmd_ptr = cmd_buffer;

    new_ptr = strchr(cmd_ptr, '\r');
    if (!new_ptr) new_ptr = cmd_buffer + CMD_BUF_SIZE;
        
    n = new_ptr - cmd_ptr; len = strlen(cmd_line);
    strncat(cmd_line, cmd_ptr, n); cmd_line[len + n] = 0;
    cmd_ptr = new_ptr;
  }

  if (*cmd_ptr == '\r') {
    cmd_ptr++;
    if (*cmd_ptr == '\n') cmd_ptr++;
  }

  return cmd_line;
}

int recv_confirm(void) {
  int rv = 0; 
  sa.sa_handler = SIG_DFL; sigaction(SIGINT, &sa, 0);

  /* here the user interrupt causes program abort.
     The problem is, if he breaks here he will lose sync with the server.
     So we reset sigint to the default handler for the moment. */

  if (recv_answer(0, 0) == 226) rv = 1;
  sa.sa_handler = &handler_INT; sigaction(SIGINT, &sa, 0);
  return rv;
}

int recv_answer(int store, struct line_data* d) {
  char* answer; int code; char str_code[] = { 0, 0, 0, ' ', 0 };

  reset_cmd_buffer();

  answer = recv_line(cmdftp_control);
  
  if ((!answer) || (strlen(answer) < 3)) return 0;

  strncpy(str_code, answer, 3);

  if (answer[3] == '-') {
    while ((answer = recv_line(cmdftp_control))) {       
      if (strncmp(answer, str_code, 4) == 0) break;
      if (store) store_line(answer, d);
      else printf("%s\n", answer);
    }
  }
  sscanf(str_code, "%i", &code);
  return code;
}

int getport() {
  unsigned int b[6]; 
  int i, answer; size_t len; char* port_str;

  if ((answer = send_command("PASV")) != 227) return 0;

  port_str = cmd_buffer + 4;
  len = strlen(port_str += 4);

  for (i = 0; i < len; i++) 
    if (!isdigit(port_str[i])) port_str[i] = ' ';
  
  if (sscanf(port_str, "%u %u %u %u %u %u", b, b + 1, b + 2, b + 3, b + 4, b + 5) != 6) return 0;

  return b[4] * 256 + b[5];
}

off_t getsize(char* filename) {
  off_t size;
  snprintf(cmd_buffer, CMD_BUF_SIZE, "SIZE %s", filename);
  if (send_command(cmd_buffer) != 213) { return -1; }
  
  sscanf(cmd_buffer + 4, OFF_FMT, &size);
  return size;
}


/*******************************/
/* MEMORY AND STRING FUNCTIONS */
/*******************************/


void* my_malloc(size_t s) {
  void* rv;
  if (!(rv = malloc(s))) cmdftp_err(CMDFTP_ERR_HEAP, "");
  return rv;
}

void* my_realloc(void* ptr, size_t s) {
  void* rv;
  if (!(rv = realloc(ptr, s))) cmdftp_err(CMDFTP_ERR_HEAP, "");
  return rv;
}

#if !HAVE_STRDUP

char* strdup(const char* s) {
  char* rv = malloc(strlen(s) + 1);
  strcpy(rv, s);
  return rv;
}

#endif

char* my_strdup(char* s) {
  char* rv;
  if (!(rv = strdup(s))) cmdftp_err(CMDFTP_ERR_HEAP, "");
  return rv;
}


/***************************/
/* OTHER UTILITY FUNCTIONS */
/***************************/


void split_cmd(char* cmd, char** argv) {
  if ((argv[1] = strchr(cmd, ' ')) == 0)
    argv[1] = cmd + strlen(cmd);
  else {
    while (*argv[1] == ' ') *argv[1]++ = 0;
  }
  argv[0] = cmd;
}

int str_binsearch(char* key) {
  int low, high, mid, chk;

  low = 0; high = N_COMMANDS - 1;

  while (low <= high) {
    mid = (low + high) / 2;
    if ((chk = strcmp(key, commands[mid])) > 0) low = mid + 1;
    else if (chk < 0) high = mid - 1;
    else return mid;
  }

  return -1;
}

char* readline(int enable_tab, int enable_echo) {
  char* ptr; int c;

  ptr = cmd_userinput; memset(cmd_userinput, 0, sizeof(cmd_userinput));
  cmdftp_raw_mode();
  
  while (1) {
    c = fgetc(stdin);
    if (c == '\n')
      { *ptr = 0; fputc('\n', stdout); break; }
    else if ((c == '\t') && (enable_tab) &&
	     (ptr - cmd_userinput) < (CMD_BUF_SIZE - 10))
      { readline_tab(); ptr = cmd_userinput + strlen(cmd_userinput); }
    else if (c == cmdftp_t.c_cc[VERASE] || c == 8)
      readline_bs(&ptr, enable_echo);
    else if ((c >= 32) && (c <= 127)) {
      if ((ptr - cmd_userinput) < (CMD_BUF_SIZE - 10))
	{ *ptr++ = c; if (enable_echo) fputc(c, stdout); }
    }
    else if (c == EOF) break;
  }
  cmdftp_canon_mode();
  return (c != EOF) ? my_strdup(cmd_userinput) : 0;
}

void readline_bs(char** ptr, int enable_echo) {
  if (*ptr > cmd_userinput) {
    *(--(*ptr)) = 0;
    if (enable_echo) fprintf(stdout, "\b \b");
  }
}

void readline_tab(void) {
  char* cmd, *first, *second, *mask, *completion; struct list_data ld;
  int c, i, j, lc_mode; size_t len, len_buffer;
  typedef int (*fetch_list_ptr)(char*, struct list_data*);
  fetch_list_ptr fetch_list[2] = { &local_fetch_list, &remote_fetch_list };

  cmd = first = second = mask = 0;
  c = i = j = len = len_buffer = 0;

  init_list(&ld);

  if ((first = strchr(cmd_userinput, ' '))) {
    strcat(cmd_userinput, "*");
    while (*first == ' ') *first++ = 0;
    if ((second = strchr(first, ' '))) {
      while (*second == ' ') *second++ = 0;
      second = my_strdup(second);
    }
    first = my_strdup(first);

  } else return;
  
  cmd = my_strdup(cmd_userinput);
  
  if ((strcmp(cmd, "d") == 0) || (strcmp(cmd, "dr") == 0)) {
    lc_mode = second ? CMDFTP_LOCAL : CMDFTP_REMOTE;
  } else if (strcmp(cmd, "u") == 0) {
    lc_mode = second ? CMDFTP_REMOTE : CMDFTP_LOCAL;
  } else lc_mode = mode;
  
  if (!(mask = second)) mask = first;
  len_buffer = strlen(mask) - 1;
  
  if ( (!fetch_list[lc_mode](mask, &ld)) || (ld.count == 0) ) goto endtab;
  len = strlen(ld.data[0].fullname);
  
  for (i = 0; i < len; i++) {
    c = ld.data[0].fullname[i];
    for (j = 1; j < ld.count; j++) {
      if (ld.data[j].fullname[i] != c) goto endcommon;
    }
  }
  
 endcommon:

 endtab:
  if (ld.count) {
    completion = my_malloc(i + 1);
    strncpy(completion, ld.data[0].fullname, i);
    completion[i] = 0;
  } else {
    completion = my_malloc(len_buffer + 1);
    strncpy(completion, mask, len_buffer);
    completion[len_buffer] = 0;
  }
  
  snprintf(cmd_userinput, CMD_BUF_SIZE, "%s%s%s%s%s", cmd,
	   first ? " " : "", second ? first : "", second ? " " : "",
	   completion);
  free(completion);
  
  if (ld.count > 1) {
    fputc('\n', stdout);
    for (j = 0; j < ld.count; j++) {
      printf("%s\n", ld.data[j].fullname);
    }
  } else {
    fputc('\r', stdout);
  }

  printf("cmdftp:%s>%s", cwd[mode], cmd_userinput); 

  if (first) free(first); if (second) free(second); if (cmd) free(cmd);
  free_list(&ld);
}

void cmdftp_pwd_start(void) {
  cmdftp_t.c_lflag &= ~ECHO; tcsetattr(0, TCSANOW, &cmdftp_t);
}

void cmdftp_pwd_end(void) {
  cmdftp_t.c_lflag |= ECHO; tcsetattr(0, TCSANOW, &cmdftp_t);
}

void cmdftp_raw_mode(void) {
  cmdftp_t.c_lflag &= ~ICANON; cmdftp_t.c_lflag &= ~ECHO;
  tcsetattr(0, TCSANOW, &cmdftp_t);
}

void cmdftp_canon_mode(void) {
  cmdftp_t.c_lflag |= ICANON; cmdftp_t.c_lflag |= ECHO;
  tcsetattr(0, TCSANOW, &cmdftp_t);
}

int cmdftp_execute(char* p1, char* p2, int read_fd, int write_fd) {
  char** argv = 0; char* argptr;
  pid_t pid; int status, rv, count;
  rv = count = 0;

  argptr = p1 = my_strdup(p1);

  do {
    argv = my_realloc(argv, sizeof(char*) * (count + 1));
    argv[count++] = argptr;
    argptr = strchr(argptr + 1, ' ');

    if (argptr) *argptr++ = 0;

  } while (argptr);

  count += (p2 ? 2 : 1); argv = my_realloc(argv, sizeof(char*) * count);
  argv[count - 1] = 0; if (p2) argv[count - 2] = p2;

  if ((pid = fork()) < 0) goto end_proc;
  else if (pid > 0) {
    if (waitpid(pid, &status, 0) > 0) {
      if (WIFEXITED(status) && (WEXITSTATUS(status) == 0)) rv = 1;
    }
  } else {
    if (read_fd != -1) { close(0); dup(read_fd); close(read_fd); }
    if (write_fd != -1) { close(1); dup(write_fd); close(write_fd); }
    execvp(argv[0], argv);
    exit(-1);
  }

 end_proc:
  free(p1); free(argv);
  return rv;
}

char cmdftp_temp_fn[] = "/tmp/cmdftpXXXXXX";

#if !HAVE_MKSTEMP

int mkstemp(char* template) {
#define MKSTEMP_ATTEMPTS 10
  int rv, attempt; size_t len; char *ptr; off_t n;
  static const char ascii[] = 
    "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

  len = strlen(template);	/* must be >= 6 */
  ptr = template + len - 6;     /* point to XXXXXX part */
  n = *((off_t*)o.hostname);	/* a crazy thing */
  if (n < 0) n = -n;

  for (attempt = 0; attempt < MKSTEMP_ATTEMPTS; attempt++) {
    int i;
    for (i = 0; i < 6; i++) {
      ptr[i] = ascii[n % (sizeof(ascii) - 1)]; n /= sizeof(ascii - 1);
    }
    rv = open(template, O_RDWR | O_CREAT | O_EXCL, 0600);
    if (rv >= 0 || errno != EEXIST) return rv;
  }
  errno = EEXIST;
  return -1;
#undef MKSTEMP_ATTEMPTS
}

#endif

FILE* cmdftp_temp(char** fn) {
  int i, d; char* ptr; FILE* rv;

  ptr = cmdftp_temp_fn + 11; *fn = 0;

  for (i = 0; i < 6; i++) ptr[i] = 'X';
  if ((d = mkstemp(cmdftp_temp_fn)) < 0) return 0;
  if (!(rv = fdopen(d, "r+"))) return 0;
  *fn = my_strdup(cmdftp_temp_fn);
  return rv;
}

void canonized_fn(char* des[3], char* arg) {
  char* tmpmask;

  tmpmask = my_strdup(arg);
  des[0] = my_strdup(dirname(tmpmask));
  free(tmpmask);

  tmpmask = my_strdup(arg);
  des[1] = my_strdup(basename(tmpmask));
  free(tmpmask);

  des[2] = fullpath(des[0], des[1]);
}

/*****************************************************************************/

void free_fn(char* des[3]) {
  int i;
  for (i = 0; i < 3; i++) free(des[i]);
}

/*****************************************************************************/

char* clean_fn(char* fn) {
  if (strncmp(fn, "./", 2) == 0) return fn + 2;
  else if(strncmp(fn, "//", 2) == 0) return fn + 1;
  else return fn;
}

/*****************************************************************************/

char* fullpath(char* s, char* q) {
  char* rv = my_malloc(strlen(s) + strlen(q) + 2);
  sprintf(rv, "%s/%s", s, q);
  return rv;
}

/********************/
/* SIGNAL FUNCTIONS */
/********************/

void init_signals(void) {
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = SIG_IGN; sigaction(SIGPIPE, &sa, 0); 
  sa.sa_handler = &handler_INT; sigaction(SIGINT, &sa, 0);
}

void handler_INT(int i) { transfer_interrupted = TRAN_INTR_INT; }


/******************************/
/* STRUCT LINE_DATA FUNCTIONS */
/******************************/


void init_lines(struct line_data* d) {
  d->count = 0; d->lines = 0;
}

void init_list(struct list_data* d) {
  d->count = 0; d->data = 0;
}

void store_line(char* line, struct line_data* d) {
  int i = d->count;
  d->lines = my_realloc(d->lines, sizeof(char*) * (++d->count));
  d->lines[i] = my_malloc(CMD_BUF_SIZE - 9);
  strncpy(d->lines[i], line, CMD_BUF_SIZE - 10);
  d->lines[i][CMD_BUF_SIZE - 10] = 0; 
}

void store_list(char* fullname, struct list_data* d) {
  int i; char* base;
  base = basename(fullname);

  if ((o.d && (strcmp(base, ".") == 0 || strcmp(base, "..") == 0)) ||
      (!o.d && *base == '.')) return;
  
  i = d->count;
  d->data = my_realloc(d->data, sizeof(struct list_entry) * (++d->count));
  d->data[i].fullname = my_strdup(fullname);
  d->data[i].basename = my_strdup(base);
  d->data[i].dirname = my_strdup(dirname(fullname));
}

void store_pretty_list(char* fullname, struct list_data* d) {
  int i = d->count;
  d->data = my_realloc(d->data, sizeof(struct list_entry) * (++d->count));
  d->data[i].fullname = my_strdup(fullname);
  d->data[i].basename = d->data[i].dirname = 0;
}

void free_lines(struct line_data* d) {
  int i;
  if (d->lines) {
    for (i = 0; i < d->count; i++) free(d->lines[i]);
    free(d->lines); d->lines = 0;
  } 
  d->count = 0;
}

void free_list(struct list_data* d) {
  int i;
  if (d->data) {
    for (i = 0; i < d->count; i++) {
      if (d->data[i].fullname) free(d->data[i].fullname);
      if (d->data[i].basename) free(d->data[i].basename);
      if (d->data[i].dirname) free(d->data[i].dirname);
    }
    free(d->data); d->data = 0;
  }
  d->count = 0;
}

