/*
           *** cmdftp, command line ftp client ***

    Copyright (C) 2003-2004 Claudio Fontana

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program (look for the file called COPYING);
    if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    You can contact the author (Claudio Fontana) by sending a mail
    to sick_soul@users.sourceforge.net
	                                                                 */

#ifndef CMDFTP_H
#define CMDFTP_H

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "auto-includes.h"

#ifndef AF_INET
#define AF_INET PF_INET
#endif

#if !HAVE_FSEEKO
# define fseeko fseek
# define ftello ftell
#endif

#if (SIZEOF_OFF_T == SIZEOF_LONG_LONG)
# define CMDFTP_LARGE_FILE 1
# define OFF_FMT "%lld"
#else
# define OFF_FMT "%ld"
#endif

#if !HAVE_STRCHR
# define strchr index
# define strrchr rindex
#endif

#if !HAVE_MEMCPY
# define memcpy(d, s, n) bcopy((s), (d), (n))
# define memset(s, c, n) bzero((s), (n))
#endif

#if !HAVE_SNPRINTF
# define snprintf(str, size, format, ...) sprintf(str, format, __VA_ARGS__)
#endif

#define CMD_BUF_SIZE 2047

enum { CMDFTP_LOCAL, CMDFTP_REMOTE };

enum { CMDFTP_ERR_NULL, CMDFTP_ERR_SERV, CMDFTP_ERR_UOPT, CMDFTP_ERR_WOPT,
       CMDFTP_ERR_NOPT, CMDFTP_ERR_CONN, CMDFTP_ERR_LGIN, CMDFTP_ERR_UNER,
       CMDFTP_ERR_HEAP };

enum { CMDFTP_WAR_BYE,
       CMDFTP_WAR_OPEN, CMDFTP_WAR_WRIT, CMDFTP_WAR_READ, CMDFTP_WAR_NOCT,
       CMDFTP_WAR_MASK, CMDFTP_WAR_TEMP, CMDFTP_WAR_BADC, CMDFTP_WAR_MARG,
       CMDFTP_WAR_IARG,
       CMDFTP_WAR_LGIN, CMDFTP_WAR_FAIL, CMDFTP_WAR_MENV, CMDFTP_WAR_LOST,
       CMDFTP_WAR_INTR };

enum { CMDFTP_ENV_PAGER, CMDFTP_ENV_EDITOR, CMDFTP_ENV_HOME, N_ENV };

enum { TRAN_INTR_NO, TRAN_INTR_INT, TRAN_INTR_PIPE };

typedef int (*fun_ptr)(char*);

struct cmdftp_options {
  char* hostname; unsigned short p, b, a, q, m, n, g, d;
};

void getoptions(int argc, char* argv[]);

struct line_data {
  int count;
  char** lines;
};

struct list_entry {
  char* fullname;
  char* basename;
  char* dirname;
};

struct list_data {
  int count;
  struct list_entry* data;
};

void init_lines(struct line_data*);
void init_list(struct list_data*);
void free_lines(struct line_data*);
void free_list(struct list_data*);

void reset_cmd_buffer(void);

void cmdftp_error(int code, char* msg);
void cmdftp_war(int code, char* msg);
void usage(void);
void version(void);

void greeting(void);

void login_procedure(void);
int login(char* user, char* pass);
int manual_login();
int auto_login();
int auto_login_next_token(char** hay, char* store);
void auto_login_read_token(char** hay, char* store);

void print_progress(off_t, off_t);
int dispatch(char* cmd);

void* my_malloc(size_t s);
void* my_realloc(void* ptr, size_t s);
char* my_strdup(char* s);

void split_cmd(char* cmd, char** argv);

int quit(char*);
int h(char*);

int local_mode(char*);
int remote_mode(char*);

int local_cd(char* arg);
int local_cd_aux(char* arg);
int local_cp(char* arg);
int local_cp_aux(char* trg, char* src);
int local_copy(char* target, struct list_entry* source);
int local_mv(char* arg);
int local_ls(char* arg);
int local_dir(char* arg);
int local_md(char* arg);
int local_rd(char* arg);
int local_rm(char* arg);
int local_p(char* arg);
int local_pwd(char* arg);
int local_e(char* arg);


int u(char* arg);
int u_aux(char* arg, char* src);
int upload(char* target, FILE* source);

char* local_pwd_aux(void);

int remote_cd(char* arg);
int remote_cd_aux(char* arg);
int remote_cp(char* arg);
int remote_cp_aux(char* trg, char* src);
int remote_copy(char* target, struct list_entry* source);
int remote_mv(char* arg);
int remote_mv_aux(char* trg, char* src);
int remote_ls(char* arg);
int remote_dir(char* arg);
int remote_md(char* arg);
int remote_rd(char* arg);
int remote_rm(char* arg);
int remote_p(char* arg);
int remote_pwd(char* arg);
int remote_e(char* arg);

int d(char* arg); int dr(char* arg);
int d_aux(char* trg, char* src, int resume);
int download(FILE* target, char* source);

char* remote_pwd_aux(void);

int local_fetch_list(char* mask, struct list_data* d);
int local_fetch_pretty_list(struct list_data* d);
int local_fetch_list_aux(char* mask, struct list_data* d, char pretty);

int remote_fetch_list(char* mask, struct list_data* d);
int remote_fetch_pretty_list(struct list_data* d);
int remote_fetch_list_aux(char* mask, struct list_data* d, char pretty);

int ls(struct list_data*);

int cmdftp_connect(int port);
void cmdftp_reconnect(void);

ssize_t my_raw_read(char* buf, size_t n, int sc);
ssize_t my_raw_write(char* buf, size_t n, int sc);

int send_command(char* cmd);
char* recv_line(int sc);
int recv_answer(int store, struct line_data* d);
int recv_confirm(void);
void store_line(char* line, struct line_data* d);
void store_list(char*, struct list_data* d);
void store_pretty_list(char*, struct list_data* d);
int getport(void);
off_t getsize(char* filename);

int str_binsearch(char* key);
char* readline(int enable_tab, int enable_echo);
void readline_tab(void);
void readline_bs(char** ptr, int enable_echo);

void cmdftp_pwd_start(void);
void cmdftp_pwd_end(void);

int cmdftp_execute(char* command, char* last_arg, int read_fd, int write_fd);
FILE* cmdftp_temp(char** fn);
void canonized_fn(char* target[3], char* source);
void free_fn(char* target[3]);
char* clean_fn(char*);
char* fullpath(char*, char*);

void handler_ALRM(int);
void handler_INT(int);
void handler_SEGV(int);

void init_signals(void);

void cmdftp_raw_mode(void);
void cmdftp_canon_mode(void);

#define TRANSFER_INTERRUPTED_CHECK(TRAN_INTR_CMD, TRAN_INTR_SYNC)  \
                                                                   \
 (transfer_interrupted == TRAN_INTR_INT) {                         \
  transfer_interrupted = TRAN_INTR_NO;                             \
  cmdftp_war(CMDFTP_WAR_INTR, TRAN_INTR_CMD);                      \
                                                                   \
  if (TRAN_INTR_SYNC) {                                            \
    sa.sa_handler = SIG_IGN; sigaction(SIGINT, &sa, 0);            \
    if (recv_answer(0, 0) == 150)                                  \
      { close(cmdftp_data); recv_answer(0, 0); }                   \
    sa.sa_handler = &handler_INT; sigaction(SIGINT, &sa, 0);       \
  }                                                                \
                                                                   \
} else if (transfer_interrupted == TRAN_INTR_PIPE) {               \
  transfer_interrupted = TRAN_INTR_NO;                             \
  cmdftp_war(CMDFTP_WAR_LOST, TRAN_INTR_CMD);                      \
  cmdftp_reconnect();                                              \
  goto start_transfer;                                             \
} else if (0)


#endif /* CMDFTP_H */
