.TH cmdftp 1 "2004-11-24" "Claudio Fontana"
.SH NAME
.B cmdftp
\- command line File Transfer Protocol client

.SH SYNOPSIS

.B cmdftp
[
.I "-a:b:dghnmp:q"
]
.I hostname

.SH DESCRIPTION

.PP
cmdftp is an alternative client for the File Transfer Protocol.

It is quite different from the default 
.BR ftp (1)
client, because it always uses passive mode for all data transfers,
has a syntax that mimics the shell, has local and remote mode, 
supports multiple and recursive file transfers, has more complex 
commands like copy and move, permits remote and local text file viewing and
editing, is able to detect network failures and resume the operation,
may resume broken downloads between sessions, has tab completion like the shell
for both local and remote file names, and now has large file support.

The ftp host to connect to must be specified on invocation ( parameter
.I hostname
). 

First of all, the client will attempt autologin, using information provided
in
.I ~/.netrc
; if it fails, or the autologin feature has been disabled (
.I "-n"
), a user/password pair must be manually entered.
By default, failure to login causes the program to quit; the number of manual
login attempts can be altered with the
.I "-a"
option.

The cmdftp:/> prompt appears as soon as the login operation completes,
and informs the user that the client is ready to receive commands: see
.B COMMANDS
section. The current directory is shown in the prompt.

The prompt does not appear if the quiet option (
.I "-q"
) has been specified.

There are many more options that control how cmdftp behaves: see 
.B OPTIONS
section for a full list and description.

Additionally, this client can interface with other programs using environment
variables: see 
.B ENVIRONMENT
section.

This program can be easily automated: see
.B USAGE_NOTES
section for tips and notes about the cmdftp netrc parser.

.SH OPTIONS

.TP 
.B "-h"
show command line options and exit.
If your system supports long options, then long option equivalents are shown.
.TP 
.B "-v"
show version information and exit
.TP
.BI "-p " FTP_PORT
specify the remote TCP port of the server (def:21)
.TP
.BI "-b " BUFFER_SIZE
specify the size in bytes of the transfer buffer (def:8192)
.TP
.BI "-a " N_ATTEMPTS
max number of manual login attempts before giving up
.TP
.B "-q "
be quiet (suppress copyright message, warnings and transfer progress)
.TP
.B "-n "
disable manual login. Rely on autologin only.
.TP
.B "-n "
disable autologin (~/.netrc). Rely on manual login only.
.TP
.B "-g "
disable output paging
.TP
.B "-d"
disable forced ignore of dot-files

.SH ENVIRONMENT
.TP
.B PAGER
refers to command p; sets the pager for file viewing and ls output. 
If the variable is not set, the output is simply not paged.
.TP
.B EDITOR
refers to command e, and is required to be able to edit local and remote files
from cmdftp. If no editor is set in the EDITOR variable, the e command will
not be available.
.TP
.B HOME
refers to the autologin feature, and is required to be able to use ~/.netrc
(to find your home directory actually). It should be already set on your
system. Note: double check ~/.netrc file permissions!

.SH COMMANDS
.TP
.B h
display this help
.TP
.B l
switch to local mode, following commands refer to local
.TP
.B r              
switch to remote mode, following commands refer to remote
.TP
.B pwd
prompt working directory
.TP
.BI "cd " PATH
change working directory
.TP
.BI "md " PATH
make new directory
.TP
.BI "rd " PATH
remove empty directory
.TP
.BI "rm " MASK
delete regular files matching MASK, skip directories
.TP
.B ls
list current directory contents
.TP
.BI "ls " MASK
list files/dirs matching MASK
.TP
.B dir
pretty list of files in current directory
.TP
.BI "cp " "MASK DIR"
copy all files/dirs matching MASK to DIR (recurs) *
.TP
.BI "mv " "MASK DIR"
move all files/dirs matching MASK to DIR (recurs) *
.TP
.BI "u " "MASK DIR"
upload files/dirs matching MASK into remote DIR (recurs) *
.TP
.BI "d " "MASK DIR"
download files/dirs matching MASK into local DIR (recurs) *
.TP
.BI "dr " "MASK DIR"
same as above, and if local file already exists, resume **
.TP
.BI "p " FILE
print contents of the FILE on the terminal
.TP
.BI "e " FILE
edit FILE
.TP
.B "q "
quit client
.TP
.B "quit|exit|bye"
aliases for q command
.TP
.B <TAB>
tab-completion for filenames

.SH COMMAND NOTES

.TP
.B *
remember that you MUST specify a target directory, even if you want to
mean "current". Use '.' (current directory) in this case.
If DIR does not exist, and MASK matches a single directory, cmdftp
tries to create a directory called DIR in which that single directory's
contents are copied.
Note that mv is currently NOT usable to rename files, only to move them
between directories! This might change in the future.

.TP
.B **
resuming has only sense between sessions, or if you explicitly interrupt
the transfer using CTRL+C or equivalent, because cmdftp will try to
resume the current command (eventually reconnecting) if any network
problems are detected. Also, using dr command might be dangerous if used
incorrecly, because if the incomplete file is a DIFFERENT file instead 
(same name, different content) the target file can be screwed up.
If you can't understand the problem just use the basic d command and you
will be fine.

.SH USAGE_NOTES
.PP
Some tips to automate cmdftp:

.TP
.B "- "
use the autologin feature and provide valid login information in 
.I ~/.netrc
: see netrc page for the full file format description. Some notes about
cmdftp netrc parser follow near the end of this section.

.TP
.B "- "
use the quiet option (
.I -q
), although redirecting stdout and stderr to a logfile could be useful instead

.TP
.B "- "
store common command sequences in a file, then run cmdftp with stdin redirected

.TP
.B "- "
use things like cron (
.BR crond (8),
.BR crontab (1)
) to execute cmdftp scripts periodically.

.PP
The cmdftp netrc parser recognizes C-like escape sequences in field values
(NOT in keys). Those are:

\\a, \\b, \\f, \\n, \\t, \\v,

\\xHH, (HH are hex digits) - use to specify a character

\\OOO, (OOO are octal digits) - use to specify a character

Escaping other characters has the effect of getting that character.
Use to escape special meaning characters, like \\ itself or the space.
Quotes have NO special meaning so don't quote your strings.
Here is an example netrc which contains some of those escape sequences:

.PD 0
.PP
machine somewhere.myhost.com
.PP
login the\\ spaced\\ user
.PP
password \\x64eadly\\040habit

.PP
machine anotherhost.hisnet.org
.PP
login with\\\\backslash
.PP
password \\tcrazy_tabs_in_pass\\t

.PP
default
.PP
login anon\\ymous
.PP
password guest

.PD
Again, see netrc man page to get the full netrc file format description.
Note that 'account' and 'macdef' are currently NOT supported.
       
.SH HOME PAGE
.I http://cmdftp.sourceforge.net/

.I http://sourceforge.net/projects/cmdftp/

.SH AUTHOR
Claudio Fontana

<sick_soul@users.sourceforge.net>

.SH FILES

.I ~/.netrc
\- autologin file

.SH SEE ALSO

.BR ftp (1),
.BR ftpd (8),
.BR netrc,
RFC 959
